

from headers import get_headers
import sys
import io
import gzip
import struct
from rpmfile import cpiofile
from functools import wraps
from rpmfile.io_extra import _SubFile

pad = lambda fileobj: (4 - (fileobj.tell() % 4)) % 4

class RPMInfo(object):
    '''
    Informational class which holds the details about an
    archive member given by an RPM entry block.
    RPMInfo objects are returned by RPMFile.getmember() and 
    RPMFile.getmembers() and are
    usually created internally.
    '''
    _new_coder = struct.Struct(b'8s8s8s8s8s8s8s8s8s8s8s8s8s')
    
    def __init__(self, name, file_start, file_size, initial_offset, isdir):
        self.name = name
        self.file_start = file_start
        self.size = file_size
        self.initial_offset = initial_offset
        self._isdir = isdir
        
    @property
    def isdir(self):
        return self._isdir
         
        
    def __repr__(self):
        return '<RPMMember %r>' % self.name
    
    @classmethod
    def _read(cls, magic, fileobj):
        if magic == '070701':
            return cls._read_new(fileobj, magic=magic)
        else:
            raise Exception('bad magic number %r' % magic)
    
    @classmethod
    def _read_new(cls, fileobj, magic=None):
        coder = cls._new_coder 
        
        initial_offset = fileobj.tell()
        d = coder.unpack_from(fileobj.read(coder.size))
        
        namesize = int(d[11], 16)
        name = fileobj.read(namesize)[:-1]
        fileobj.seek(pad(fileobj), 1)
        file_start = fileobj.tell()
        file_size = int(d[6], 16)
        fileobj.seek(file_size, 1)
        fileobj.seek(pad(fileobj), 1)
        nlink = int(d[4], 16)
        isdir = nlink == 2 and file_size == 0  
        return cls(name, file_start, file_size, initial_offset, isdir)
    
class RPMFile(object):
    '''
    Open an RPM archive `name'. `mode' must be 'r' to
    read from an existing archive.
    
    If `fileobj' is given, it is used for reading or writing data. If it
    can be determined, `mode' is overridden by `fileobj's mode.
    `fileobj' is not closed, when TarFile is closed.

    '''
    def __init__(self, name=None, mode='rb', fileobj=None):
        
        if mode != 'rb':
            raise NotImplementedError("currently the only supported mode is 'rb'")
        self._fileobj = fileobj or io.open(name, mode)
        self._headers = get_headers(self._fileobj)
        self.data_offset = self._fileobj.tell()
        self._ownes_fd = fileobj is not None
        
    @property
    def headers(self):
        'RPM headers'
        return self._headers
    
    def __enter__(self):
        return self
    
    def __exit__(self, *excinfo):
        if self._ownes_fd:
            self._fileobj.close()
        
    _members = None
    
    def getmembers(self):
        '''
        Return the members of the archive as a list of RPMInfo objects. The
        list has the same order as the members in the archive.
        '''
        if self._members is None:
            self._members = _members = []
            g = self.gzip_file
            magic = g.read(2)
            while magic: 
                if magic == '07':
                    magic += g.read(4)
                    member = RPMInfo._read(magic, g)
                    
                    if member.name == 'TRAILER!!!':
                        break
                    
                    if not member.isdir:
                        _members.append(member)
                    
                magic = g.read(2)
            return _members
        return self._members
    
    def getmember(self, name):
        '''
        Return an RPMInfo object for member `name'. If `name' can not be
        found in the archive, KeyError is raised. If a member occurs more
        than once in the archive, its last occurrence is assumed to be the
        most up-to-date version.
        '''
        members = self.getmembers()
        for m in members[::-1]:
            if m.name == name:
                return m
            
        raise KeyError("member %s could not be found" % name)
        
    def extractfile(self, member):
        '''
        Extract a member from the archive as a file object. `member' may be
        a filename or an RPMInfo object.
        The file-like object is read-only and provides the following
        methods: read(), readline(), readlines(), seek() and tell()
        '''
        if not isinstance(member, RPMInfo):
            member = self.getmember(member)
        return _SubFile(self.gzip_file, member.file_start, member.size)

    _gzip_file = None
    
    @property
    def gzip_file(self):
        'Return the uncompressed raw CPIO data of the RPM archive'
        if self._gzip_file is None:
            fileobj = _SubFile(self._fileobj, self.data_offset)
            self._gzip_file = gzip.GzipFile(fileobj=fileobj)
        return self._gzip_file
    
def open(name=None, mode='rb', fileobj=None):
    '''
    Open an RPM archive for reading. Return
    an appropriate RPMFile class.
    '''
    return RPMFile(name, mode, fileobj)

def main():
    print sys.argv[1]
    with open(sys.argv[1]) as rpm:
        print rpm.headers
        for m in rpm.getmembers():
            print m 
        print 'done'
            

if __name__ == '__main__':
    main()



'''
Created on Jan 10, 2014

@author: sean
'''
import sys
import struct
from rpmfile.errors import RPMError
import re
import io
import rpmdefs

RPM_LEAD_MAGIC_NUMBER = '\xed\xab\xee\xdb'
RPM_HEADER_MAGIC_NUMBER = '\x8e\xad\xe8'
HEADER_MAGIC_NUMBER = re.compile('(\x8e\xad\xe8)')

RPMTAGS = {rpmdefs.RPMTAG_NAME: 'name',
           rpmdefs.RPMTAG_VERSION: 'version',
           rpmdefs.RPMTAG_RELEASE: 'release',
           rpmdefs.RPMTAG_DESCRIPTION: 'description',
           rpmdefs.RPMTAG_COPYRIGHT: 'copyright',
           rpmdefs.RPMTAG_URL: 'url',
           rpmdefs.RPMTAG_ARCH: 'arch'}

class Entry(object):
    ''' RPM Header Entry
    '''
    def __init__(self, entry, store):
        self.entry = entry
        self.store = store

        self.switch = { rpmdefs.RPM_DATA_TYPE_CHAR:            self.__readchar,
                        rpmdefs.RPM_DATA_TYPE_INT8:            self.__readint8,
                        rpmdefs.RPM_DATA_TYPE_INT16:           self.__readint16,
                        rpmdefs.RPM_DATA_TYPE_INT32:           self.__readint32,
                        rpmdefs.RPM_DATA_TYPE_INT64:           self.__readint64,
                        rpmdefs.RPM_DATA_TYPE_STRING:          self.__readstring,
                        rpmdefs.RPM_DATA_TYPE_BIN:             self.__readbin,
                        rpmdefs.RPM_DATA_TYPE_I18NSTRING_TYPE: self.__readstring}

        self.store.seek(entry[2])
        self.value = self.switch[entry[1]]()
        self.tag = entry[0]

    def __str__(self):
        return "(%s, %s)" % (self.tag, self.value, )

    def __repr__(self):
        return "(%s, %s)" % (self.tag, self.value, )

    def __readchar(self, offset=1):
        ''' store is a pointer to the store offset
        where the char should be read
        '''
        data = self.store.read(offset)
        fmt = '!'+str(offset)+'c'
        value = struct.unpack(fmt, data)
        return value

    def __readint8(self, offset=1):
        ''' int8 = 1byte
        '''
        return self.__readchar(offset)

    def __readint16(self, offset=1):
        ''' int16 = 2bytes
        '''
        data = self.store.read(offset*2)
        fmt = '!'+str(offset)+'i'
        value = struct.unpack(fmt, data)
        return value

    def __readint32(self, offset=1):
        ''' int32 = 4bytes
        '''
        data = self.store.read(offset*4)
        fmt = '!'+str(offset)+'i'
        value = struct.unpack(fmt, data)
        return value

    def __readint64(self, offset=1):
        ''' int64 = 8bytes
        '''
        data = self.store.read(offset*4)
        fmt = '!'+str(offset)+'l'
        value = struct.unpack(fmt, data)
        return value

    def __readstring(self):
        ''' read a string entry
        '''
        string = ''
        while True:
            char = self.__readchar()
            if char[0] == '\x00': # read until '\0'
                break
            string += char[0]
        return string

    def __readbin(self):
        ''' read a binary entry
        '''
        if self.entry[0] == rpmdefs.RPMSIGTAG_MD5:
            data = self.store.read(rpmdefs.MD5_SIZE)
            value = struct.unpack('!'+rpmdefs.MD5_SIZE+'s', data)
            return value
        elif self.entry[0] == rpmdefs.RPMSIGTAG_PGP:
            data = self.store.read(rpmdefs.PGP_SIZE)
            value = struct.unpack('!'+rpmdefs.PGP_SIZE+'s', data)
            return value

class Header(object):
    ''' RPM Header Structure
    '''
    def __init__(self, header, entries , store):
        '''
        '''
        self.header = header
        self.entries = entries
        self.store = store
        self.pentries = []
        self.rentries = []

        self.__readentries()


    def __readentry(self, entry):
        ''' [4bytes][4bytes][4bytes][4bytes]
               TAG    TYPE   OFFSET  COUNT
        '''
        entryfmt = '!llll'
        entry = struct.unpack(entryfmt, entry)
        if entry[0] < rpmdefs.RPMTAG_MIN_NUMBER or\
                entry[0] > rpmdefs.RPMTAG_MAX_NUMBER:
            return None
        return entry

    def __readentries(self):
        ''' read a rpm entry
        '''
        for entry in self.entries:
            entry = self.__readentry(entry)
            if entry:
                if entry[0] in rpmdefs.RPMTAGS:
                    self.pentries.append(entry)

        for pentry in self.pentries:
            entry = Entry(pentry, self.store)
            if entry:
                self.rentries.append(entry)


def _readheader(header):
    ''' reads the header-header section
    [3bytes][1byte][4bytes][4bytes][4bytes]
      MN      VER   UNUSED  IDXNUM  STSIZE
    '''
    headerfmt = '!3sc4sll'
    if not len(header) == 16:
        raise RPMError('invalid header size')

    header = struct.unpack(headerfmt, header)
    magic_num = header[0]
    if magic_num != RPM_HEADER_MAGIC_NUMBER:
        raise RPMError('invalid RPM header')
    return header


def _readheaders(fileobj, offset):
    ''' read information headers
    '''
    # lets find the start of the header
    fileobj.seek(offset)
    start = find_magic_number(HEADER_MAGIC_NUMBER, fileobj)
    # go back to the begining of the header
    fileobj.seek(start)
    header = fileobj.read(16)
    header = _readheader(header)
    entries = []
    _headers = []
    for entry in range(header[3]):
        _entry = fileobj.read(16)
        entries.append(_entry)
    store = io.BytesIO(fileobj.read(header[4]))
    _headers.append(Header(header, entries, store))

    _entries = {}
    for header in _headers:
        for entry in header.rentries:
            key = RPMTAGS.get(entry.tag, entry.tag)
            _entries[key] = entry.value 
    return _entries

def find_magic_number(regexp, data):
    ''' find a magic number in a buffer
    '''
    string = data.read(1)
    while True:
        match = regexp.search(string)
        if match:
            return data.tell() - 3
        byte = data.read(1)
        if not byte:
            return None
        else:
            string += byte

def _readlead(fileobj):
    ''' reads the rpm lead section

        struct rpmlead {
           unsigned char magic[4];
           unsigned char major, minor;
           short type;
           short archnum;
           char name[66];
           short osnum;
           short signature_type;
           char reserved[16];
           } ;
    '''
    lead_fmt = '!4sBBhh66shh16s'
    data = fileobj.read(96)
    value = struct.unpack(lead_fmt, data)

    magic_num = value[0]
    ptype = value[3]

    if magic_num != RPM_LEAD_MAGIC_NUMBER:
        raise RPMError('wrong magic number this is not a RPM file')

    if ptype not in [0, 1]:
        raise RPMError('wrong package type this is not a RPM file')

def _read_sigheader(fileobj):
    ''' read signature header

        ATN: this will not return any usefull information
        besides the file offset
    '''
    start = find_magic_number(HEADER_MAGIC_NUMBER, fileobj)
    if not start:
        raise RPMError('invalid RPM file, signature header not found')
    # return the offsite after the magic number
    return start + 3

def get_headers(fileobj):
    _readlead(fileobj)
    offset = _read_sigheader(fileobj)
    return _readheaders(fileobj, offset)
    
def main():
    with open(sys.argv[1]) as fileobj:
        print get_headers(fileobj)

if __name__ == '__main__':
    main() 



# -*- coding: iso-8859-15 -*-
# -*- Mode: Python; py-ident-offset: 4 -*-
# vim:ts=4:sw=4:et
'''
rpm definitions

$Id$
'''
__revision__ = '$Rev$'[6:-2]

RPM_LEAD_MAGIC_NUMBER = '\xed\xab\xee\xdb'
RPM_HEADER_MAGIC_NUMBER = '\x8e\xad\xe8'

RPMTAG_MIN_NUMBER = 1000
RPMTAG_MAX_NUMBER = 1146

#signature tags
RPMSIGTAG_SIZE     = 1000
RPMSIGTAG_LEMD5_1  = 1001
RPMSIGTAG_PGP      = 1002
RPMSIGTAG_LEMD5_2  = 1003
RPMSIGTAG_MD5      = 1004
RPMSIGTAG_GPG      = 1005
RPMSIGTAG_PGP5     = 1006


MD5_SIZE = 16 #16 bytes long
PGP_SIZE = 152 #152 bytes long


# data types definition
RPM_DATA_TYPE_NULL = 0
RPM_DATA_TYPE_CHAR = 1
RPM_DATA_TYPE_INT8 = 2
RPM_DATA_TYPE_INT16 = 3
RPM_DATA_TYPE_INT32 = 4
RPM_DATA_TYPE_INT64 = 5
RPM_DATA_TYPE_STRING = 6
RPM_DATA_TYPE_BIN = 7
RPM_DATA_TYPE_STRING_ARRAY = 8
RPM_DATA_TYPE_I18NSTRING_TYPE = 9

RPM_DATA_TYPES = (RPM_DATA_TYPE_NULL,
                  RPM_DATA_TYPE_CHAR,
                  RPM_DATA_TYPE_INT8,
                  RPM_DATA_TYPE_INT16,
                  RPM_DATA_TYPE_INT32,
                  RPM_DATA_TYPE_INT64,
                  RPM_DATA_TYPE_STRING,
                  RPM_DATA_TYPE_BIN,
                  RPM_DATA_TYPE_STRING_ARRAY,)

RPMTAG_NAME = 1000
RPMTAG_VERSION = 1001
RPMTAG_RELEASE = 1002
RPMTAG_DESCRIPTION = 1005
RPMTAG_COPYRIGHT = 1014
RPMTAG_URL = 1020
RPMTAG_ARCH = 1022


RPMTAGS = (RPMTAG_NAME,
           RPMTAG_VERSION,
           RPMTAG_RELEASE,
           RPMTAG_DESCRIPTION,
           RPMTAG_COPYRIGHT,
           RPMTAG_URL,
           RPMTAG_ARCH,)



'''
@author: sean
'''

from setuptools import setup, find_packages

setup(
    name='rpmfile',
    description='Read rmp archive files',
    version="0.1",
    author='Sean Ross-Ross',
    author_email='srossross@gmail.com',
    packages=find_packages(),

)
